@@ -2,7 +2,7 @@
 /*                                                                       */
 /*    This file is part of the HiGHS linear optimization suite           */
 /*                                                                       */
-/*    Written and engineered 2008-2023 by Julian Hall, Ivet Galabova,    */
+/*    Written and engineered 2008-2024 by Julian Hall, Ivet Galabova,    */
 /*    Leona Gottwald and Michael Feldmeier                               */
 /*                                                                       */
 /*    Available as open-source under the MIT License                     */
@@ -29,6 +29,8 @@
 
 using std::fabs;
 
+const bool report_mip_timing = false;
+
 HighsMipSolver::HighsMipSolver(HighsCallback& callback,
                                const HighsOptions& options, const HighsLp& lp,
                                const HighsSolution& solution, bool submip)
@@ -61,9 +63,8 @@ HighsMipSolver::HighsMipSolver(HighsCallback& callback,
       obj += orig_model_->col_cost_[i] * value;
 
       if (orig_model_->integrality_[i] == HighsVarType::kInteger) {
-        double intval = std::floor(value + 0.5);
         integrality_violation_ =
-            std::max(fabs(intval - value), integrality_violation_);
+            std::max(fractionality(value), integrality_violation_);
       }
 
       const double lower = orig_model_->col_lower_[i];
@@ -104,7 +105,8 @@ HighsMipSolver::~HighsMipSolver() = default;
 
 void HighsMipSolver::run() {
   modelstatus_ = HighsModelStatus::kNotset;
-  // std::cout << options_mip_->presolve << std::endl;
+  // Start the solve_clock for the timer that is local to the HighsMipSolver
+  // instance
   timer_.start(timer_.solve_clock);
   improving_solution_file_ = nullptr;
   if (!submip && options_mip_->mip_improving_solution_file != "")
@@ -113,7 +115,16 @@ void HighsMipSolver::run() {
 
   mipdata_ = decltype(mipdata_)(new HighsMipSolverData(*this));
   mipdata_->init();
-  mipdata_->runPresolve();
+  mipdata_->runPresolve(options_mip_->presolve_reduction_limit);
+  if (report_mip_timing & !submip)
+    highsLogUser(options_mip_->log_options, HighsLogType::kInfo,
+                 "MIP-Timing: After %6.4fs - completed mipdata_->runPresolve\n",
+                 timer_.read(timer_.solve_clock));
+  // Identify whether time limit has been reached (in presolve)
+  if (modelstatus_ == HighsModelStatus::kNotset &&
+      timer_.read(timer_.solve_clock) >= options_mip_->time_limit)
+    modelstatus_ = HighsModelStatus::kTimeLimit;
+
   if (modelstatus_ != HighsModelStatus::kNotset) {
     highsLogUser(options_mip_->log_options, HighsLogType::kInfo,
                  "Presolve: %s\n",
@@ -121,16 +132,40 @@ void HighsMipSolver::run() {
     if (modelstatus_ == HighsModelStatus::kOptimal) {
       mipdata_->lower_bound = 0;
       mipdata_->upper_bound = 0;
-      mipdata_->transformNewIncumbent(std::vector<double>());
+      mipdata_->transformNewIntegerFeasibleSolution(std::vector<double>());
+      mipdata_->saveReportMipSolution();
     }
     cleanupSolve();
     return;
   }
 
+  if (report_mip_timing & !submip)
+    highsLogUser(options_mip_->log_options, HighsLogType::kInfo,
+                 "MIP-Timing: After %6.4fs - reached   mipdata_->runSetup()\n",
+                 timer_.read(timer_.solve_clock));
   mipdata_->runSetup();
+  if (report_mip_timing & !submip)
+    highsLogUser(options_mip_->log_options, HighsLogType::kInfo,
+                 "MIP-Timing: After %6.4fs - completed mipdata_->runSetup()\n",
+                 timer_.read(timer_.solve_clock));
 restart:
   if (modelstatus_ == HighsModelStatus::kNotset) {
+    // Check limits have not been reached before evaluating root node
+    if (mipdata_->checkLimits()) {
+      cleanupSolve();
+      return;
+    }
+    if (report_mip_timing & !submip)
+      highsLogUser(
+          options_mip_->log_options, HighsLogType::kInfo,
+          "MIP-Timing: After %6.4fs - reached   mipdata_->evaluateRootNode()\n",
+          timer_.read(timer_.solve_clock));
     mipdata_->evaluateRootNode();
+    if (report_mip_timing & !submip)
+      highsLogUser(
+          options_mip_->log_options, HighsLogType::kInfo,
+          "MIP-Timing: After %6.4fs - completed mipdata_->evaluateRootNode()\n",
+          timer_.read(timer_.solve_clock));
     // age 5 times to remove stored but never violated cuts after root
     // separation
     mipdata_->cutpool.performAging();
@@ -139,7 +174,7 @@ restart:
     mipdata_->cutpool.performAging();
     mipdata_->cutpool.performAging();
   }
-  if (mipdata_->nodequeue.empty()) {
+  if (mipdata_->nodequeue.empty() || mipdata_->checkLimits()) {
     cleanupSolve();
     return;
   }
@@ -173,7 +208,7 @@ restart:
     HighsInt iterlimit = 10 * std::max(mipdata_->lp.getAvgSolveIters(),
                                        mipdata_->avgrootlpiters);
     iterlimit = std::max({HighsInt{10000}, iterlimit,
-                          HighsInt(1.5 * mipdata_->firstrootlpiters)});
+                          HighsInt((3 * mipdata_->firstrootlpiters) / 2)});
 
     mipdata_->lp.setIterationLimit(iterlimit);
 
@@ -235,7 +270,7 @@ restart:
 
       search.flushStatistics();
       mipdata_->printDisplayLine();
-      // printf("continue plunging due to good esitmate\n");
+      // printf("continue plunging due to good estimate\n");
     }
     search.openNodesToQueue(mipdata_->nodequeue);
     search.flushStatistics();
@@ -335,13 +370,19 @@ restart:
           lowerBoundLastCheck = mipdata_->lower_bound;
         }
 
-        int64_t minHugeTreeOffset =
-            (mipdata_->num_leaves - mipdata_->num_leaves_before_run) * 1e-3;
-        int64_t minHugeTreeEstim = HighsIntegers::nearestInteger(
-            activeIntegerRatio * (10 + minHugeTreeOffset) *
-            std::pow(1.5, nTreeRestarts));
+        // Possibly prevent restart - necessary for debugging presolve
+        // errors: see #1553
+        if (options_mip_->mip_allow_restart) {
+          int64_t minHugeTreeOffset =
+              (mipdata_->num_leaves - mipdata_->num_leaves_before_run) / 1000;
+          int64_t minHugeTreeEstim = HighsIntegers::nearestInteger(
+              activeIntegerRatio * (10 + minHugeTreeOffset) *
+              std::pow(1.5, nTreeRestarts));
 
-        doRestart = numHugeTreeEstim >= minHugeTreeEstim;
+          doRestart = numHugeTreeEstim >= minHugeTreeEstim;
+        } else {
+          doRestart = false;
+        }
       } else {
         // count restart due to many fixings within the first 1000 nodes as
         // root restart
@@ -484,6 +525,7 @@ restart:
 }
 
 void HighsMipSolver::cleanupSolve() {
+  mipdata_->printDisplayLine('Z');
   timer_.start(timer_.postsolve_clock);
   bool havesolution = solution_objective_ != kHighsInf;
   bool feasible;
@@ -544,14 +586,13 @@ void HighsMipSolver::cleanupSolve() {
   else
     gap_ = kHighsInf;
 
-  std::array<char, 128> gapString;
+  std::array<char, 128> gapString = {};
 
   if (gap_ == kHighsInf)
     std::strcpy(gapString.data(), "inf");
   else {
     double printTol = std::max(std::min(1e-2, 1e-1 * gap_), 1e-6);
-    std::array<char, 32> gapValString =
-        highsDoubleToString(100.0 * gap_, printTol);
+    auto gapValString = highsDoubleToString(100.0 * gap_, printTol);
     double gapTol = options_mip_->mip_rel_gap;
 
     if (options_mip_->mip_abs_gap > options_mip_->mip_feasibility_tolerance) {
@@ -566,8 +607,7 @@ void HighsMipSolver::cleanupSolve() {
                     gapValString.data());
     else if (gapTol != kHighsInf) {
       printTol = std::max(std::min(1e-2, 1e-1 * gapTol), 1e-6);
-      std::array<char, 32> gapTolString =
-          highsDoubleToString(100.0 * gapTol, printTol);
+      auto gapTolString = highsDoubleToString(100.0 * gapTol, printTol);
       std::snprintf(gapString.data(), gapString.size(),
                     "%s%% (tolerance: %s%%)", gapValString.data(),
                     gapTolString.data());
@@ -614,10 +654,14 @@ void HighsMipSolver::cleanupSolve() {
   assert(modelstatus_ != HighsModelStatus::kNotset);
 }
 
-void HighsMipSolver::runPresolve() {
+void HighsMipSolver::runPresolve(const HighsInt presolve_reduction_limit) {
+  // Start the solve_clock for the timer that is local to the HighsMipSolver
+  // instance
+  assert(!timer_.running(timer_.solve_clock));
+  timer_.start(timer_.solve_clock);
   mipdata_ = decltype(mipdata_)(new HighsMipSolverData(*this));
   mipdata_->init();
-  mipdata_->runPresolve();
+  mipdata_->runPresolve(presolve_reduction_limit);
 }
 
 const HighsLp& HighsMipSolver::getPresolvedModel() const {
@@ -631,3 +675,26 @@ HighsPresolveStatus HighsMipSolver::getPresolveStatus() const {
 presolve::HighsPostsolveStack HighsMipSolver::getPostsolveStack() const {
   return mipdata_->postSolveStack;
 }
+
+void HighsMipSolver::callbackGetCutPool() const {
+  assert(callback_->user_callback);
+  assert(callback_->callbackActive(kCallbackMipGetCutPool));
+  HighsCallbackDataOut& data_out = callback_->data_out;
+
+  std::vector<double> cut_lower;
+  std::vector<double> cut_upper;
+  HighsSparseMatrix cut_matrix;
+
+  mipdata_->lp.getCutPool(data_out.cutpool_num_col, data_out.cutpool_num_cut,
+                          cut_lower, cut_upper, cut_matrix);
+
+  data_out.cutpool_num_nz = cut_matrix.numNz();
+  data_out.cutpool_start = cut_matrix.start_.data();
+  data_out.cutpool_index = cut_matrix.index_.data();
+  data_out.cutpool_value = cut_matrix.value_.data();
+  data_out.cutpool_lower = cut_lower.data();
+  data_out.cutpool_upper = cut_upper.data();
+  callback_->user_callback(kCallbackMipGetCutPool, "MIP cut pool",
+                           &callback_->data_out, &callback_->data_in,
+                           callback_->user_callback_data);
+}
