@@ -2,7 +2,7 @@
 /*                                                                       */
 /*    This file is part of the HiGHS linear optimization suite           */
 /*                                                                       */
-/*    Written and engineered 2008-2023 by Julian Hall, Ivet Galabova,    */
+/*    Written and engineered 2008-2024 by Julian Hall, Ivet Galabova,    */
 /*    Leona Gottwald and Michael Feldmeier                               */
 /*                                                                       */
 /*    Available as open-source under the MIT License                     */
@@ -15,6 +15,7 @@
 #include "lp_data/HConst.h"
 #include "lp_data/HighsOptions.h"
 #include "util/HighsCDouble.h"
+#include "util/HighsUtils.h"
 
 namespace presolve {
 
@@ -72,12 +73,25 @@ void HighsPostsolveStack::LinearTransform::transformToPresolvedSpace(
   primalSol[col] /= scale;
 }
 
+static HighsBasisStatus computeRowStatus(double dual,
+                                         HighsPostsolveStack::RowType rowType) {
+  if (rowType == HighsPostsolveStack::RowType::kEq)
+    return dual < 0 ? HighsBasisStatus::kUpper : HighsBasisStatus::kLower;
+  else if (rowType == HighsPostsolveStack::RowType::kGeq)
+    return HighsBasisStatus::kLower;
+  else
+    return HighsBasisStatus::kUpper;
+}
+
 void HighsPostsolveStack::FreeColSubstitution::undo(
     const HighsOptions& options, const std::vector<Nonzero>& rowValues,
     const std::vector<Nonzero>& colValues, HighsSolution& solution,
     HighsBasis& basis) {
-  double colCoef = 0;
+  // a (removed) cut may have been used in this reduction.
+  bool isModelRow = static_cast<size_t>(row) < solution.row_value.size();
+
   // compute primal values
+  double colCoef = 0;
   HighsCDouble rowValue = 0;
   for (const auto& rowVal : rowValues) {
     if (rowVal.index == col)
@@ -88,34 +102,53 @@ void HighsPostsolveStack::FreeColSubstitution::undo(
 
   assert(colCoef != 0);
   // Row values aren't fully postsolved, so why do this?
-  solution.row_value[row] =
-      double(rowValue + colCoef * solution.col_value[col]);
+  if (isModelRow)
+    solution.row_value[row] =
+        double(rowValue + colCoef * solution.col_value[col]);
   solution.col_value[col] = double((rhs - rowValue) / colCoef);
 
   // if no dual values requested, return here
   if (!solution.dual_valid) return;
 
   // compute the row dual value such that reduced cost of basic column is 0
-  solution.row_dual[row] = 0;
-  HighsCDouble dualval = colCost;
-  for (const auto& colVal : colValues)
-    dualval -= colVal.value * solution.row_dual[colVal.index];
+  if (isModelRow) {
+    solution.row_dual[row] = 0;
+    HighsCDouble dualval = colCost;
+    for (const auto& colVal : colValues) {
+      if (static_cast<size_t>(colVal.index) < solution.row_dual.size())
+        dualval -= colVal.value * solution.row_dual[colVal.index];
+    }
+    solution.row_dual[row] = double(dualval / colCoef);
+  }
 
   solution.col_dual[col] = 0;
-  solution.row_dual[row] = double(dualval / colCoef);
 
   // set basis status if necessary
   if (!basis.valid) return;
 
   basis.col_status[col] = HighsBasisStatus::kBasic;
-  if (rowType == RowType::kEq)
-    basis.row_status[row] = solution.row_dual[row] < 0
-                                ? HighsBasisStatus::kUpper
-                                : HighsBasisStatus::kLower;
-  else if (rowType == RowType::kGeq)
-    basis.row_status[row] = HighsBasisStatus::kLower;
+  if (isModelRow)
+    basis.row_status[row] = computeRowStatus(solution.row_dual[row], rowType);
+}
+
+static HighsBasisStatus computeStatus(double dual, HighsBasisStatus& status,
+                                      double dual_feasibility_tolerance) {
+  if (dual > dual_feasibility_tolerance)
+    status = HighsBasisStatus::kLower;
+  else if (dual < -dual_feasibility_tolerance)
+    status = HighsBasisStatus::kUpper;
+
+  return status;
+}
+
+static HighsBasisStatus computeStatus(double dual,
+                                      double dual_feasibility_tolerance) {
+  if (dual > dual_feasibility_tolerance)
+    return HighsBasisStatus::kLower;
+  else if (dual < -dual_feasibility_tolerance)
+    return HighsBasisStatus::kUpper;
   else
-    basis.row_status[row] = HighsBasisStatus::kUpper;
+    return HighsBasisStatus::kBasic;
 }
 
 void HighsPostsolveStack::DoubletonEquation::undo(
@@ -129,39 +162,34 @@ void HighsPostsolveStack::DoubletonEquation::undo(
   // can only do primal postsolve, stop here
   if (row == -1 || !solution.dual_valid) return;
 
-  HighsBasisStatus colStatus;
+  const HighsBasisStatus colStatus =
+      !basis.valid
+          ? computeStatus(solution.col_dual[col],
+                          options.dual_feasibility_tolerance)
+          : computeStatus(solution.col_dual[col], basis.col_status[col],
+                          options.dual_feasibility_tolerance);
 
-  if (basis.valid) {
-    if (solution.col_dual[col] > options.dual_feasibility_tolerance)
-      basis.col_status[col] = HighsBasisStatus::kLower;
-    else if (solution.col_dual[col] < -options.dual_feasibility_tolerance)
-      basis.col_status[col] = HighsBasisStatus::kUpper;
-
-    colStatus = basis.col_status[col];
-  } else {
-    if (solution.col_dual[col] > options.dual_feasibility_tolerance)
-      colStatus = HighsBasisStatus::kLower;
-    else if (solution.col_dual[col] < -options.dual_feasibility_tolerance)
-      colStatus = HighsBasisStatus::kUpper;
-    else
-      colStatus = HighsBasisStatus::kBasic;
-  }
+  // a (removed) cut may have been used in this reduction.
+  bool isModelRow = static_cast<size_t>(row) < solution.row_value.size();
 
   // compute the current dual values of the row and the substituted column
   // before deciding on which column becomes basic
-  // for each entry in a row i of the substituted column we added the doubleton
-  // equation row with scale -a_i/substCoef. Therefore the dual multiplier of
-  // this row i implicitly increases the dual multiplier of this doubleton
-  // equation row with that scale.
+  // for each entry in a row i of the substituted column we added the
+  // doubleton equation row with scale -a_i/substCoef. Therefore the dual
+  // multiplier of this row i implicitly increases the dual multiplier of this
+  // doubleton equation row with that scale.
   HighsCDouble rowDual = 0.0;
-  solution.row_dual[row] = 0;
-  for (const auto& colVal : colValues)
-    rowDual -= colVal.value * solution.row_dual[colVal.index];
-  rowDual /= coefSubst;
-  solution.row_dual[row] = double(rowDual);
-
-  // the equation was also added to the objective, so the current values need to
-  // be changed
+  if (isModelRow) {
+    solution.row_dual[row] = 0;
+    for (const auto& colVal : colValues) {
+      if (static_cast<size_t>(colVal.index) < solution.row_dual.size())
+        rowDual -= colVal.value * solution.row_dual[colVal.index];
+    }
+    rowDual /= coefSubst;
+    solution.row_dual[row] = double(rowDual);
+  }
+  // the equation was also added to the objective, so the current values need
+  // to be changed
   solution.col_dual[colSubst] = substCost;
   solution.col_dual[col] += substCost * coef / coefSubst;
 
@@ -171,7 +199,7 @@ void HighsPostsolveStack::DoubletonEquation::undo(
     // so alter the dual multiplier of the row to make the dual multiplier of
     // column zero
     double rowDualDelta = solution.col_dual[col] / coef;
-    solution.row_dual[row] = double(rowDual + rowDualDelta);
+    if (isModelRow) solution.row_dual[row] = double(rowDual + rowDualDelta);
     solution.col_dual[col] = 0.0;
     solution.col_dual[colSubst] = double(
         HighsCDouble(solution.col_dual[colSubst]) - rowDualDelta * coefSubst);
@@ -187,10 +215,10 @@ void HighsPostsolveStack::DoubletonEquation::undo(
       basis.col_status[col] = HighsBasisStatus::kBasic;
     }
   } else {
-    // otherwise make the reduced cost of the subsituted column zero and make
+    // otherwise make the reduced cost of the substituted column zero and make
     // that column basic
     double rowDualDelta = solution.col_dual[colSubst] / coefSubst;
-    solution.row_dual[row] = double(rowDual + rowDualDelta);
+    if (isModelRow) solution.row_dual[row] = double(rowDual + rowDualDelta);
     solution.col_dual[colSubst] = 0.0;
     solution.col_dual[col] =
         double(HighsCDouble(solution.col_dual[col]) - rowDualDelta * coef);
@@ -199,15 +227,18 @@ void HighsPostsolveStack::DoubletonEquation::undo(
 
   if (!basis.valid) return;
 
-  if (solution.row_dual[row] < 0)
-    basis.row_status[row] = HighsBasisStatus::kLower;
-  else
-    basis.row_status[row] = HighsBasisStatus::kUpper;
+  if (isModelRow)
+    basis.row_status[row] = computeRowStatus(solution.row_dual[row], rowType);
 }
 
 void HighsPostsolveStack::EqualityRowAddition::undo(
     const HighsOptions& options, const std::vector<Nonzero>& eqRowValues,
     HighsSolution& solution, HighsBasis& basis) const {
+  // (removed) cuts may have been used in this reduction.
+  if (static_cast<size_t>(row) >= solution.row_value.size() ||
+      static_cast<size_t>(addedEqRow) >= solution.row_value.size())
+    return;
+
   // nothing more to do if the row is zero in the dual solution or there is
   // no dual solution
   if (!solution.dual_valid || solution.row_dual[row] == 0.0) return;
@@ -225,18 +256,22 @@ void HighsPostsolveStack::EqualityRowAdditions::undo(
     const HighsOptions& options, const std::vector<Nonzero>& eqRowValues,
     const std::vector<Nonzero>& targetRows, HighsSolution& solution,
     HighsBasis& basis) const {
+  // a (removed) cut may have been used in this reduction.
+  if (static_cast<size_t>(addedEqRow) >= solution.row_value.size()) return;
+
   // nothing more to do if the row is zero in the dual solution or there is
   // no dual solution
   if (!solution.dual_valid) return;
 
   // the dual multiplier of the rows where the eq row was added implicitly
-  // increases the dual multiplier of the equation with the scale that was used
-  // for adding the equation
+  // increases the dual multiplier of the equation with the scale that was
+  // used for adding the equation
   HighsCDouble eqRowDual = solution.row_dual[addedEqRow];
-  for (const auto& targetRow : targetRows)
-    eqRowDual +=
-        HighsCDouble(targetRow.value) * solution.row_dual[targetRow.index];
-
+  for (const auto& targetRow : targetRows) {
+    if (static_cast<size_t>(targetRow.index) < solution.row_dual.size())
+      eqRowDual +=
+          HighsCDouble(targetRow.value) * solution.row_dual[targetRow.index];
+  }
   solution.row_dual[addedEqRow] = double(eqRowDual);
 
   assert(!basis.valid);
@@ -251,32 +286,36 @@ void HighsPostsolveStack::ForcingColumn::undo(
 
   HighsInt debug_num_use_row_value = 0;
   const bool debug_report = false;
+
+  auto computeColVal = [&](HighsInt direction) {
+    // choose solution value that makes all rows feasible
+    for (const auto& colVal : colValues) {
+      // Row values aren't fully postsolved, so how can this work?
+      debug_num_use_row_value++;
+      if (static_cast<size_t>(colVal.index) < solution.row_value.size()) {
+        double colValFromRow = solution.row_value[colVal.index] / colVal.value;
+        if (direction * colValFromRow > direction * colValFromNonbasicRow) {
+          nonbasicRow = colVal.index;
+          colValFromNonbasicRow = colValFromRow;
+          nonbasicRowStatus = direction * colVal.value > 0
+                                  ? HighsBasisStatus::kLower
+                                  : HighsBasisStatus::kUpper;
+        }
+      }
+    }
+    // round solution value if column is integer-constrained
+    if (nonbasicRow != -1 && colIntegral)
+      colValFromNonbasicRow =
+          direction * std::ceil(direction * colValFromNonbasicRow -
+                                options.mip_feasibility_tolerance);
+  };
+
   if (atInfiniteUpper) {
     // choose largest value as then all rows are feasible
-    for (const auto& colVal : colValues) {
-      // Row values aren't fully postsolved, so how can this work?
-      debug_num_use_row_value++;
-      double colValFromRow = solution.row_value[colVal.index] / colVal.value;
-      if (colValFromRow > colValFromNonbasicRow) {
-        nonbasicRow = colVal.index;
-        colValFromNonbasicRow = colValFromRow;
-        nonbasicRowStatus = colVal.value > 0 ? HighsBasisStatus::kLower
-                                             : HighsBasisStatus::kUpper;
-      }
-    }
+    computeColVal(HighsInt{1});
   } else {
     // choose smallest value, as then all rows are feasible
-    for (const auto& colVal : colValues) {
-      // Row values aren't fully postsolved, so how can this work?
-      debug_num_use_row_value++;
-      double colValFromRow = solution.row_value[colVal.index] / colVal.value;
-      if (colValFromRow < colValFromNonbasicRow) {
-        nonbasicRow = colVal.index;
-        colValFromNonbasicRow = colValFromRow;
-        nonbasicRowStatus = colVal.value > 0 ? HighsBasisStatus::kUpper
-                                             : HighsBasisStatus::kLower;
-      }
-    }
+    computeColVal(HighsInt{-1});
   }
   if (debug_num_use_row_value && debug_report) {
     printf(
@@ -305,8 +344,11 @@ void HighsPostsolveStack::ForcingColumn::undo(
 void HighsPostsolveStack::ForcingColumnRemovedRow::undo(
     const HighsOptions& options, const std::vector<Nonzero>& rowValues,
     HighsSolution& solution, HighsBasis& basis) const {
-  // we use the row value as storage for the scaled value implied on the column
-  // dual
+  // a (removed) cut may have been used in this reduction.
+  if (static_cast<size_t>(row) >= solution.row_value.size()) return;
+
+  // we use the row value as storage for the scaled value implied on the
+  // column dual
   HighsCDouble val = rhs;
   for (const auto& rowVal : rowValues)
     val -= rowVal.value * solution.col_value[rowVal.index];
@@ -321,65 +363,61 @@ void HighsPostsolveStack::ForcingColumnRemovedRow::undo(
 void HighsPostsolveStack::SingletonRow::undo(const HighsOptions& options,
                                              HighsSolution& solution,
                                              HighsBasis& basis) const {
+  // a (removed) cut may have been used in this reduction.
+  bool isModelRow = static_cast<size_t>(row) < solution.row_value.size();
+
   // nothing to do if the rows dual value is zero in the dual solution or
   // there is no dual solution
   if (!solution.dual_valid) return;
 
-  HighsBasisStatus colStatus;
-
-  if (basis.valid) {
-    if (solution.col_dual[col] > options.dual_feasibility_tolerance)
-      basis.col_status[col] = HighsBasisStatus::kLower;
-    else if (solution.col_dual[col] < -options.dual_feasibility_tolerance)
-      basis.col_status[col] = HighsBasisStatus::kUpper;
-
-    colStatus = basis.col_status[col];
-  } else {
-    if (solution.col_dual[col] > options.dual_feasibility_tolerance)
-      colStatus = HighsBasisStatus::kLower;
-    else if (solution.col_dual[col] < -options.dual_feasibility_tolerance)
-      colStatus = HighsBasisStatus::kUpper;
-    else
-      colStatus = HighsBasisStatus::kBasic;
-  }
+  const HighsBasisStatus colStatus =
+      !basis.valid
+          ? computeStatus(solution.col_dual[col],
+                          options.dual_feasibility_tolerance)
+          : computeStatus(solution.col_dual[col], basis.col_status[col],
+                          options.dual_feasibility_tolerance);
 
   if ((!colLowerTightened || colStatus != HighsBasisStatus::kLower) &&
       (!colUpperTightened || colStatus != HighsBasisStatus::kUpper)) {
     // the tightened bound is not used in the basic solution
     // hence we simply make the row basic and give it a dual multiplier of 0
-    if (basis.valid) basis.row_status[row] = HighsBasisStatus::kBasic;
-    solution.row_dual[row] = 0;
+    if (isModelRow) {
+      if (basis.valid) basis.row_status[row] = HighsBasisStatus::kBasic;
+      solution.row_dual[row] = 0;
+    }
     return;
   }
 
   // choose the row dual value such that the columns reduced cost becomes
   // zero
-  solution.row_dual[row] = solution.col_dual[col] / coef;
+  if (isModelRow) solution.row_dual[row] = solution.col_dual[col] / coef;
   solution.col_dual[col] = 0;
 
   if (!basis.valid) return;
 
-  switch (colStatus) {
-    case HighsBasisStatus::kLower:
-      assert(colLowerTightened);
-      if (coef > 0)
-        // tightened lower bound comes from row lower bound
-        basis.row_status[row] = HighsBasisStatus::kLower;
-      else
-        // tightened lower bound comes from row upper bound
-        basis.row_status[row] = HighsBasisStatus::kUpper;
+  if (isModelRow) {
+    switch (colStatus) {
+      case HighsBasisStatus::kLower:
+        assert(colLowerTightened);
+        if (coef > 0)
+          // tightened lower bound comes from row lower bound
+          basis.row_status[row] = HighsBasisStatus::kLower;
+        else
+          // tightened lower bound comes from row upper bound
+          basis.row_status[row] = HighsBasisStatus::kUpper;
 
-      break;
-    case HighsBasisStatus::kUpper:
-      if (coef > 0)
-        // tightened upper bound comes from row lower bound
-        basis.row_status[row] = HighsBasisStatus::kUpper;
-      else
-        // tightened lower bound comes from row upper bound
-        basis.row_status[row] = HighsBasisStatus::kLower;
-      break;
-    default:
-      assert(false);
+        break;
+      case HighsBasisStatus::kUpper:
+        if (coef > 0)
+          // tightened upper bound comes from row lower bound
+          basis.row_status[row] = HighsBasisStatus::kUpper;
+        else
+          // tightened lower bound comes from row upper bound
+          basis.row_status[row] = HighsBasisStatus::kLower;
+        break;
+      default:
+        assert(false);
+    }
   }
 
   // column becomes basic
@@ -400,8 +438,8 @@ void HighsPostsolveStack::FixedCol::undo(const HighsOptions& options,
 
   HighsCDouble reducedCost = colCost;
   for (const auto& colVal : colValues) {
-    assert((HighsInt)solution.row_dual.size() > colVal.index);
-    reducedCost -= colVal.value * solution.row_dual[colVal.index];
+    if (static_cast<size_t>(colVal.index) < solution.row_dual.size())
+      reducedCost -= colVal.value * solution.row_dual[colVal.index];
   }
 
   solution.col_dual[col] = double(reducedCost);
@@ -419,6 +457,9 @@ void HighsPostsolveStack::FixedCol::undo(const HighsOptions& options,
 void HighsPostsolveStack::RedundantRow::undo(const HighsOptions& options,
                                              HighsSolution& solution,
                                              HighsBasis& basis) const {
+  // a (removed) cut may have been used in this reduction.
+  if (static_cast<size_t>(row) >= solution.row_value.size()) return;
+
   // set row dual to zero if dual solution requested
   if (!solution.dual_valid) return;
 
@@ -435,34 +476,21 @@ void HighsPostsolveStack::ForcingRow::undo(
   // compute the row dual multiplier and determine the new basic column
   HighsInt basicCol = -1;
   double dualDelta = 0;
-  if (rowType == RowType::kLeq) {
-    for (const auto& rowVal : rowValues) {
-      double colDual =
-          solution.col_dual[rowVal.index] - rowVal.value * dualDelta;
-      if (colDual * rowVal.value < 0) {
-        // column is dual infeasible, decrease the row dual such that its
-        // reduced cost become zero and remember this column as the new basic
-        // column for this row
-        dualDelta = solution.col_dual[rowVal.index] / rowVal.value;
-        basicCol = rowVal.index;
-      }
-    }
-  } else {
-    for (const auto& rowVal : rowValues) {
-      double colDual =
-          solution.col_dual[rowVal.index] - rowVal.value * dualDelta;
-      if (colDual * rowVal.value > 0) {
-        // column is dual infeasible, decrease the row dual such that its
-        // reduced cost become zero and remember this column as the new basic
-        // column for this row
-        dualDelta = solution.col_dual[rowVal.index] / rowVal.value;
-        basicCol = rowVal.index;
-      }
+  HighsInt direction = rowType == RowType::kLeq ? 1 : -1;
+  for (const auto& rowVal : rowValues) {
+    double colDual = solution.col_dual[rowVal.index] - rowVal.value * dualDelta;
+    if (direction * colDual * rowVal.value < 0) {
+      // column is dual infeasible, decrease the row dual such that its
+      // reduced cost become zero and remember this column as the new basic
+      // column for this row
+      dualDelta = solution.col_dual[rowVal.index] / rowVal.value;
+      basicCol = rowVal.index;
     }
   }
 
   if (basicCol != -1) {
-    solution.row_dual[row] = solution.row_dual[row] + dualDelta;
+    bool isModelRow = static_cast<size_t>(row) < solution.row_dual.size();
+    if (isModelRow) solution.row_dual[row] = solution.row_dual[row] + dualDelta;
     for (const auto& rowVal : rowValues) {
       solution.col_dual[rowVal.index] =
           double(solution.col_dual[rowVal.index] -
@@ -471,9 +499,10 @@ void HighsPostsolveStack::ForcingRow::undo(
     solution.col_dual[basicCol] = 0;
 
     if (basis.valid) {
-      basis.row_status[row] =
-          (rowType == RowType::kGeq ? HighsBasisStatus::kLower
-                                    : HighsBasisStatus::kUpper);
+      if (isModelRow)
+        basis.row_status[row] =
+            (rowType == RowType::kGeq ? HighsBasisStatus::kLower
+                                      : HighsBasisStatus::kUpper);
 
       basis.col_status[basicCol] = HighsBasisStatus::kBasic;
     }
@@ -483,32 +512,50 @@ void HighsPostsolveStack::ForcingRow::undo(
 void HighsPostsolveStack::DuplicateRow::undo(const HighsOptions& options,
                                              HighsSolution& solution,
                                              HighsBasis& basis) const {
+  // (removed) cuts may have been used in this reduction.
+  if (static_cast<size_t>(row) >= solution.row_value.size()) return;
+  bool duplicateIsModelRow =
+      static_cast<size_t>(duplicateRow) < solution.row_value.size();
+
   if (!solution.dual_valid) return;
   if (!rowUpperTightened && !rowLowerTightened) {
     // simple case of row2 being redundant, in which case it just gets a
     // dual multiplier of 0 and is made basic
-    solution.row_dual[duplicateRow] = 0.0;
-    if (basis.valid) basis.row_status[duplicateRow] = HighsBasisStatus::kBasic;
+    if (duplicateIsModelRow) {
+      solution.row_dual[duplicateRow] = 0.0;
+      if (basis.valid)
+        basis.row_status[duplicateRow] = HighsBasisStatus::kBasic;
+    }
     return;
   }
 
-  HighsBasisStatus rowStatus;
+  const HighsBasisStatus rowStatus =
+      !basis.valid
+          ? computeStatus(solution.row_dual[row],
+                          options.dual_feasibility_tolerance)
+          : computeStatus(solution.row_dual[row], basis.row_status[row],
+                          options.dual_feasibility_tolerance);
 
-  if (basis.valid) {
-    if (solution.row_dual[row] < -options.dual_feasibility_tolerance)
-      basis.row_status[row] = HighsBasisStatus::kUpper;
-    else if (solution.row_dual[row] > options.dual_feasibility_tolerance)
-      basis.row_status[row] = HighsBasisStatus::kLower;
-
-    rowStatus = basis.row_status[row];
-  } else {
-    if (solution.row_dual[row] < -options.dual_feasibility_tolerance)
-      rowStatus = HighsBasisStatus::kUpper;
-    else if (solution.row_dual[row] > options.dual_feasibility_tolerance)
-      rowStatus = HighsBasisStatus::kLower;
-    else
-      rowStatus = HighsBasisStatus::kBasic;
-  }
+  auto computeRowDualAndStatus = [&](bool tighened) {
+    if (tighened) {
+      if (duplicateIsModelRow) {
+        solution.row_dual[duplicateRow] =
+            solution.row_dual[row] / duplicateRowScale;
+        if (basis.valid) {
+          if (duplicateRowScale > 0)
+            basis.row_status[duplicateRow] = HighsBasisStatus::kUpper;
+          else
+            basis.row_status[duplicateRow] = HighsBasisStatus::kLower;
+        }
+      }
+      solution.row_dual[row] = 0.0;
+      if (basis.valid) basis.row_status[row] = HighsBasisStatus::kBasic;
+    } else if (duplicateIsModelRow) {
+      solution.row_dual[duplicateRow] = 0.0;
+      if (basis.valid)
+        basis.row_status[duplicateRow] = HighsBasisStatus::kBasic;
+    }
+  };
 
   // at least one bound of the row was tightened by using the bound of the
   // scaled parallel row, hence we might need to make the parallel row
@@ -517,48 +564,20 @@ void HighsPostsolveStack::DuplicateRow::undo(const HighsOptions& options,
   switch (rowStatus) {
     case HighsBasisStatus::kBasic:
       // if row is basic the parallel row is also basic
-      solution.row_dual[duplicateRow] = 0.0;
-      if (basis.valid)
-        basis.row_status[duplicateRow] = HighsBasisStatus::kBasic;
+      if (duplicateIsModelRow) {
+        solution.row_dual[duplicateRow] = 0.0;
+        if (basis.valid)
+          basis.row_status[duplicateRow] = HighsBasisStatus::kBasic;
+      }
       break;
     case HighsBasisStatus::kUpper:
       // if row sits on its upper bound, and the row upper bound was
       // tightened using the parallel row we make the row basic and
       // transfer its dual value to the parallel row with the proper scale
-      if (rowUpperTightened) {
-        solution.row_dual[duplicateRow] =
-            solution.row_dual[row] / duplicateRowScale;
-        solution.row_dual[row] = 0.0;
-        if (basis.valid) {
-          basis.row_status[row] = HighsBasisStatus::kBasic;
-          if (duplicateRowScale > 0)
-            basis.row_status[duplicateRow] = HighsBasisStatus::kUpper;
-          else
-            basis.row_status[duplicateRow] = HighsBasisStatus::kLower;
-        }
-      } else {
-        solution.row_dual[duplicateRow] = 0.0;
-        if (basis.valid)
-          basis.row_status[duplicateRow] = HighsBasisStatus::kBasic;
-      }
+      computeRowDualAndStatus(rowUpperTightened);
       break;
     case HighsBasisStatus::kLower:
-      if (rowLowerTightened) {
-        solution.row_dual[duplicateRow] =
-            solution.row_dual[row] / duplicateRowScale;
-        solution.row_dual[row] = 0.0;
-        if (basis.valid) {
-          basis.row_status[row] = HighsBasisStatus::kBasic;
-          if (duplicateRowScale > 0)
-            basis.row_status[duplicateRow] = HighsBasisStatus::kUpper;
-          else
-            basis.row_status[duplicateRow] = HighsBasisStatus::kLower;
-        }
-      } else {
-        solution.row_dual[duplicateRow] = 0.0;
-        if (basis.valid)
-          basis.row_status[duplicateRow] = HighsBasisStatus::kBasic;
-      }
+      computeRowDualAndStatus(rowLowerTightened);
       break;
     default:
       assert(false);
@@ -710,8 +729,7 @@ void HighsPostsolveStack::DuplicateColumn::undo(const HighsOptions& options,
   } else if (duplicateColIntegral) {
     // Doesn't set basis.col_status[duplicateCol], so assume no basis
     assert(!basis.valid);
-    double roundVal = std::round(solution.col_value[duplicateCol]);
-    if (std::abs(roundVal - solution.col_value[duplicateCol]) >
+    if (fractionality(solution.col_value[duplicateCol]) >
         options.mip_feasibility_tolerance) {
       solution.col_value[duplicateCol] =
           std::floor(solution.col_value[duplicateCol]);
@@ -828,7 +846,8 @@ void HighsPostsolveStack::DuplicateColumn::undo(const HighsOptions& options,
         basis.col_status[col] = HighsBasisStatus::kNonbasic;
         if (debug_report)
           printf(
-              "When demerging, neither col nor duplicateCol can be nonbasic\n");
+              "When demerging, neither col nor duplicateCol can be "
+              "nonbasic\n");
         if (kAllowDeveloperAssert) assert(666 == 999);
       }
     }
@@ -856,7 +875,7 @@ bool HighsPostsolveStack::DuplicateColumn::okMerge(
   // value of x. Hence a must be an integer and a <= (x_u-x_l)+1
   //
   // For example, if x and y are binary, then x+a.y is [0, 1, a,
-  // 1+a]. For this to be a continuous sequernce of integers, we must
+  // 1+a]. For this to be a continuous sequence of integers, we must
   // have a <= 2.
   //
   // * If x is integer and y is continuous:
@@ -886,10 +905,12 @@ bool HighsPostsolveStack::DuplicateColumn::okMerge(
   const double scale = colScale;
   const bool x_int = colIntegral;
   const bool y_int = duplicateColIntegral;
-  const double x_lo = x_int ? std::ceil(colLower) : colLower;
-  const double x_up = x_int ? std::floor(colUpper) : colUpper;
-  const double y_lo = y_int ? std::ceil(duplicateColLower) : duplicateColLower;
-  const double y_up = y_int ? std::floor(duplicateColUpper) : duplicateColUpper;
+  const double x_lo = x_int ? std::ceil(colLower - tolerance) : colLower;
+  const double x_up = x_int ? std::floor(colUpper + tolerance) : colUpper;
+  const double y_lo =
+      y_int ? std::ceil(duplicateColLower - tolerance) : duplicateColLower;
+  const double y_up =
+      y_int ? std::floor(duplicateColUpper + tolerance) : duplicateColUpper;
   const double x_len = x_up - x_lo;
   const double y_len = y_up - y_lo;
   std::string newline = "\n";
@@ -906,9 +927,7 @@ bool HighsPostsolveStack::DuplicateColumn::okMerge(
   if (x_int) {
     if (y_int) {
       // Scale must be integer and not exceed (x_u-x_l)+1 in magnitude
-      double int_scale = std::floor(scale + 0.5);
-      bool scale_is_int = std::fabs(int_scale - scale) <= tolerance;
-      if (!scale_is_int) {
+      if (fractionality(scale) > tolerance) {
         if (debug_report)
           printf(
               "%sDuplicateColumn::checkMerge: scale must be integer, but is "
@@ -962,7 +981,8 @@ bool HighsPostsolveStack::DuplicateColumn::okMerge(
       if (abs_scale > scale_limit) {
         if (debug_report)
           printf(
-              "%sDuplicateColumn::checkMerge: scale = %g, but |scale| must be "
+              "%sDuplicateColumn::checkMerge: scale = %g, but |scale| must "
+              "be "
               "at "
               "most %g since x is [%g, %g]\n",
               newline.c_str(), scale, scale_limit, x_lo, x_up);
@@ -990,14 +1010,12 @@ void HighsPostsolveStack::DuplicateColumn::undoFix(
   //=============================================================================================
 
   auto isInteger = [&](const double v) {
-    double int_v = std::floor(v + 0.5);
-    return std::fabs(int_v - v) <= mip_feasibility_tolerance;
+    return (fractionality(v) <= mip_feasibility_tolerance);
   };
 
   auto isFeasible = [&](const double l, const double v, const double u) {
-    if (v < l - primal_feasibility_tolerance) return false;
-    if (v > u + primal_feasibility_tolerance) return false;
-    return true;
+    return v >= l - primal_feasibility_tolerance &&
+           v <= u + primal_feasibility_tolerance;
   };
   const double merge_value = col_value[col];
   const double value_max = 1000;
@@ -1007,10 +1025,16 @@ void HighsPostsolveStack::DuplicateColumn::undoFix(
   const bool y_int = duplicateColIntegral;
   const int x_ix = col;
   const int y_ix = duplicateCol;
-  const double x_lo = x_int ? std::ceil(colLower) : colLower;
-  const double x_up = x_int ? std::floor(colUpper) : colUpper;
-  const double y_lo = y_int ? std::ceil(duplicateColLower) : duplicateColLower;
-  const double y_up = y_int ? std::floor(duplicateColUpper) : duplicateColUpper;
+  const double x_lo =
+      x_int ? std::ceil(colLower - mip_feasibility_tolerance) : colLower;
+  const double x_up =
+      x_int ? std::floor(colUpper + mip_feasibility_tolerance) : colUpper;
+  const double y_lo =
+      y_int ? std::ceil(duplicateColLower - mip_feasibility_tolerance)
+            : duplicateColLower;
+  const double y_up =
+      y_int ? std::floor(duplicateColUpper + mip_feasibility_tolerance)
+            : duplicateColUpper;
   if (kAllowDeveloperAssert) assert(scale);
   double x_v = merge_value;
   double y_v;
@@ -1284,14 +1308,16 @@ void HighsPostsolveStack::DuplicateColumn::undoFix(
   if (!check) {
     if (debug_report)
       printf(
-          "DuplicateColumn::undo error: std::fabs(x_v) < kHighsInf is false\n");
+          "DuplicateColumn::undo error: std::fabs(x_v) < kHighsInf is "
+          "false\n");
     if (allow_assert) assert(check);
   }
   check = std::fabs(y_v) < kHighsInf;
   if (!check) {
     if (debug_report)
       printf(
-          "DuplicateColumn::undo error: std::fabs(y_v) < kHighsInf is false\n");
+          "DuplicateColumn::undo error: std::fabs(y_v) < kHighsInf is "
+          "false\n");
     if (allow_assert) assert(check);
   }
   check = residual <= residual_tolerance;
