@@ -2,7 +2,7 @@
 /*                                                                       */
 /*    This file is part of the HiGHS linear optimization suite           */
 /*                                                                       */
-/*    Written and engineered 2008-2023 by Julian Hall, Ivet Galabova,    */
+/*    Written and engineered 2008-2024 by Julian Hall, Ivet Galabova,    */
 /*    Leona Gottwald and Michael Feldmeier                               */
 /*                                                                       */
 /*    Available as open-source under the MIT License                     */
@@ -333,6 +333,7 @@ void HighsSimplexAnalysis::iterationReport() {
   iterationReport(false);
 }
 
+// Called externally - from HEkkPrimal/Dual::reportRebuild
 void HighsSimplexAnalysis::invertReport() {
   if (*log_options.log_dev_level) {
     const bool header = (num_invert_report_since_last_header < 0) ||
@@ -373,6 +374,7 @@ void HighsSimplexAnalysis::invertReport(const bool header) {
   if (!header) num_invert_report_since_last_header++;
 }
 
+// Called externally - from HEkk::returnFromSolve
 void HighsSimplexAnalysis::userInvertReport(const bool force) {
   if (last_user_log_time < 0) {
     const bool header = true;
@@ -949,7 +951,7 @@ void HighsSimplexAnalysis::summaryReport() {
     if (num_correct_dual_primal_flip) {
       printf(
           "%12" HIGHSINT_FORMAT
-          "   correct dual primal flips (max = %g) for min dual infeasiblity "
+          "   correct dual primal flips (max = %g) for min dual infeasibility "
           "= %g\n",
           num_correct_dual_primal_flip, max_correct_dual_primal_flip,
           min_correct_dual_primal_flip_dual_infeasibility);
@@ -957,7 +959,7 @@ void HighsSimplexAnalysis::summaryReport() {
     if (num_correct_dual_cost_shift) {
       printf(
           "%12" HIGHSINT_FORMAT
-          "   correct dual  cost shifts (max = %g) for max dual infeasiblity "
+          "   correct dual  cost shifts (max = %g) for max dual infeasibility "
           "= %g\n",
           num_correct_dual_cost_shift, max_correct_dual_cost_shift,
           max_correct_dual_cost_shift_dual_infeasibility);
@@ -1090,7 +1092,6 @@ void HighsSimplexAnalysis::summaryReport() {
       double dlTime = toTime - fmTime;
       HighsInt iterSpeed = 0;
       if (dlTime > 0) iterSpeed = dlIter / dlTime;
-      HighsInt lc_simplex_strategy = lcAnIter.AnIterTrace_simplex_strategy;
       HighsInt lc_edge_weight_mode = lcAnIter.AnIterTrace_edge_weight_mode;
       std::string str_edge_weight_mode;
       if (lc_edge_weight_mode == (HighsInt)EdgeWeightMode::kSteepestEdge)
@@ -1112,9 +1113,8 @@ void HighsSimplexAnalysis::summaryReport() {
       printOneDensity(lcAnIter.AnIterTraceDensity[kSimplexNlaFtran]);
       printOneDensity(lcAnIter.AnIterTraceDensity[kSimplexNlaBtranEp]);
       printOneDensity(lcAnIter.AnIterTraceDensity[kSimplexNlaPriceAp]);
-      double use_DSE_density;
-      HighsInt local_simplex_strategy = lcAnIter.AnIterTrace_simplex_strategy;
       if (rp_dual_steepest_edge) {
+        double use_DSE_density;
         if (lc_edge_weight_mode == (HighsInt)EdgeWeightMode::kSteepestEdge) {
           use_DSE_density = lcAnIter.AnIterTraceDensity[kSimplexNlaFtranDse];
         } else {
@@ -1480,7 +1480,11 @@ void HighsSimplexAnalysis::reportIterationData(const bool header) {
 void HighsSimplexAnalysis::reportRunTime(const bool header,
                                          const double run_time) {
   if (header) return;
-  *analysis_log << highsFormatToString(" %ds", (int)run_time);
+#ifndef NDEBUG
+  *analysis_log << highsFormatToString(" %15.8gs", run_time);
+#else
+  *analysis_log << highsFormatToString(" %ds", int(run_time + 0.49));
+#endif
 }
 
 HighsInt HighsSimplexAnalysis::intLog10(const double v) {
